# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from cubicweb.devtools.testlib import CubicWebTC

from cubes.container.testutils import ContainerMixinTC


class OwnedByHookTC(ContainerMixinTC, CubicWebTC):

    def setup_database(self):
        self.create_user(self.request(), 'u')
        self.commit()
        with self.login('u'):
            req = self.request()
            self.eid = req.create_entity('CollaborativeEntity').eid
            self.commit()

    def logins_with_perm_on(self, perm, eid, req=None):
        """Logins of users with `perm` permisssion on entity with `eid`"""
        req = req or self.request()
        return [u.login for u in req.entity_from_eid(eid).related(
            'can_' + perm, role='object', entities=True)]

    def test_creator(self):
        """Creator can read/write her entity"""
        self.assertEqual(['u'], self.logins_with_perm_on('read', self.eid))
        self.assertEqual(['u'], self.logins_with_perm_on('write', self.eid))

    def test_owners(self):
        """Add/remove an owner, check read/write permissions get updated"""
        req = self.request()
        user2 = self.create_user(req, 'v')
        self.commit()
        req.execute('SET X owned_by U WHERE U login "v", X eid %d' % self.eid)
        self.commit()
        self.assertEqual(['u', 'v'],
                         self.logins_with_perm_on('read', self.eid))
        self.assertEqual(['u', 'v'],
                         self.logins_with_perm_on('write', self.eid))
        req.execute('DELETE X owned_by U WHERE U login "v", X eid %d' % self.eid)
        self.commit()
        self.assertEqual(['u'], self.logins_with_perm_on('read', self.eid))
        self.assertEqual(['u'], self.logins_with_perm_on('write', self.eid))


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# -*- coding: utf-8 -*-
# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-collaboration entity's classes"""

from cubicweb.view import EntityAdapter
from cubicweb.entities.adapters import ITreeAdapter
from cubicweb.web.views.ibreadcrumbs import IBreadCrumbsAdapter
from cubicweb.predicates import adaptable, relation_possible

from cubes.container.entities import Container, ContainerClone

COLLABORATIVE_SELECTOR = relation_possible('collaborates_on', role='object')


class ClonableAdapter(ContainerClone):
    """ The concrete adapter needs to define the selector according to the
    clone rdef.

    See CollaborationConfiguration.build_collaboration_entities.
    """
    __abstract__ = True

    def clone(self, original=None):
        super(ClonableAdapter, self).clone(original)
        self.set_meta_relations()

    def set_meta_relations(self):
        """Set meta relations for entities in the cloned container.

        `owned_by` and `created_by` relations are not handled by
        cubicweb-container (see http://www.cubicweb.org/ticket/3589525).
        """
        creator = self.entity.creator
        owners = self.entity.owned_by
        rtype = self.entity.container_config.rtype
        for row in self.entity.related(rtype, role='object').rows:
            eid = row[0]
            for rel in ('owned_by', 'created_by'):
                self._cw.execute(
                    'DELETE X %s U WHERE X eid %%(x)s' % rel, {'x': eid})
                self._cw.execute(
                    'SET X %s U WHERE X eid %%(x)s, U eid %%(u)s' % rel,
                    {'x': eid, 'u': owners[0].eid})


class CollaborativeAdapter(EntityAdapter):
    __regid__ = 'ICollaborative'
    __select__ = COLLABORATIVE_SELECTOR

    @property
    def clones(self):
        rtype, role = self.entity.cw_adapt_to('Container.clone').clone_rtype_role
        role = {'subject': 'object', 'object': 'subject'}[role]
        return self.entity.related(rtype, role=role, entities=True)

    @property
    def collaborators(self):
        return self.entity.reverse_collaborates_on

    def set_collaborators(self, collab_list):
        self._cw.execute('DELETE X collaborates_on Y WHERE EXISTS(X collaborates_on Y, Y eid %(oeid)s)',
                         {'oeid': self.entity.eid})
        self.entity.cw_set(reverse_collaborates_on=collab_list)


class CollaborativeSecurityAdapter(EntityAdapter):
    __regid__ = 'ICollaborationSecurity'
    __select__ = (COLLABORATIVE_SELECTOR &
                  relation_possible('can_write', role='object'))

    def grant_permission(self, ueid, permission):
        """ Grants users can_read / can_write permissions on entities """
        assert permission in ('read', 'write')
        if permission == 'write':
            self._cw.execute('SET U can_write W WHERE '
                             'U eid %(u)s, W eid %(w)s, NOT U can_write W',
                             {'u': ueid, 'w': self.entity.eid})
        self._cw.execute('SET U can_read W WHERE '
                         'U eid %(u)s, W eid %(w)s, NOT U can_read W',
                         {'u': ueid, 'w': self.entity.eid})

    def revoke_permission(self, ueid, permission):
        """ Revokes users can_read / can_write permissions on entities """
        assert permission in ('read', 'write')
        self._cw.execute('DELETE U can_%s W WHERE U eid %%(u)s, W eid %%(w)s'
                         % permission, {'u': ueid, 'w': self.entity.eid})


class CollaborativeITreeAdapter(ITreeAdapter):
    """The concrete adapter needs to set the `tree_relation` attribute as well
    as the selector according to the clone relation.

    See CollaborationConfiguration.build_collaboration_entities.
    """
    __abstract__ = True
    parent_role = 'object'
    child_role = 'subject'


class CollaborativeIBreadCrumbsAdapter(IBreadCrumbsAdapter):
    """For collaborative entities, do not put all the path in the collaborative
    tree in the breadcrumbs. Thus, provide a ``parent_entity`` method which
    doesn't rely on the collaboration tree (``ITree``)'s parent method.

    """

    __select__ = IBreadCrumbsAdapter.__select__ & adaptable('ICollaborative')

    def parent_entity(self):
        """ Return nothing, so that in the breadcrumbs one only displays the
        link to the current entity and to the rset of entities of the same
        type.
        """
        return None



# -*- coding: utf-8 -*-
# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-collaboration views/forms/actions/components for web ui"""
import json

from logilab.mtconverter import xml_escape

from cubicweb import _
from cubicweb.predicates import is_instance
from cubicweb.web.views.treeview import TreeViewItemView
from cubicweb.web.views.baseviews import OneLineView
from cubicweb.web.views.actions import CopyAction


from cubicweb.predicates import (adaptable, one_line_rset, has_permission,
                                 score_entity, rql_condition,
                                 match_user_groups, is_in_state)
from cubicweb.web import action, Redirect, component
from cubicweb.view import EntityView
from cubicweb.web.views import primary, uicfg, editcontroller

_pvs = uicfg.primaryview_section

_pvs.tag_object_of(('CWUser', 'collaborates_on', '*'), 'attributes')


def cloned(entity):
    adapted = entity.cw_adapt_to('ICollaborative')
    if adapted is not None:
        return 1 if adapted.clones else 0
    return 0


def not_cloned(entity):
    adapted = entity.cw_adapt_to('ICollaborative')
    if adapted is not None:
        return 0 if adapted.clones else 1
    return 0


def clonable(entity):
    """Return 1 if the entity is clonable according to its workflow state"""
    collabcfg = entity.container_config.collaboration_config
    clone_state = collabcfg.clone_state
    return is_in_state(clone_state).score_entity(entity)


class CloneAction(CopyAction):
    """Just a copy action (copy is handled by edit controller below) named 'clone'."""
    __select__ = (CopyAction.__select__ & one_line_rset() &
                  adaptable('ICollaborative') & score_entity(clonable) &
                  has_permission('add'))
    title = _('clone')

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        rtype, role = entity.cw_adapt_to('Container.clone').clone_rtype_role
        linkto = '%s:%s:%s' % (rtype, entity.eid, role)
        return entity.absolute_url(vid='copy', __linkto=linkto)


CopyAction.__select__ &= ~adaptable('ICollaborative')


class CollaborativeCloneTreeView(component.EntityCtxComponent):
    context = 'left'
    __regid__ = 'collaboration.treeview-ctx'
    order = 31
    title = _('collaborative tree')
    __select__ = (component.EntityCtxComponent.__select__ &
                  adaptable('ICollaborative'))

    def render_body(self, w):
        root = self.entity.cw_adapt_to('ITree').root()
        root.view('treeview', w=w, subvid='collaboration.current-entity',
                  current_entity=self.entity.eid)


class CurrentCollaborativeEntityView(EntityView):
    __regid__ = 'collaboration.current-entity'
    __select__ = (EntityView.__select__ &
                  adaptable('ICollaborative'))

    def cell_call(self, row, col, **morekwargs):
        entity = self.cw_rset.get_entity(row, col)
        current_entity = morekwargs.get('current_entity')
        self.w(self.html(entity, current_entity))

    def creator_html(self, entity):
        creator = entity.created_by[0]
        if creator.eid == self._cw.user.eid:
            return u''
        else:
            return u'<span class="creator">%s</span>' % creator.view('oneline')

    def entity_html(self, entity, current_entity):
        if entity.eid == current_entity:
            return xml_escape(entity.dc_title())
        else:
            return entity.view('oneline')

    def html(self, entity, current_entity):
        html = self.entity_html(entity, current_entity)
        creator_html = self.creator_html(entity)
        if creator_html:
            html += u' (%s)' % creator_html
        return html


class CollaborativeTreeItemView(TreeViewItemView):
    __select__ = (TreeViewItemView.__select__ &
                  adaptable('ICollaborative'))
    default_branch_state_is_open = True



# -*- coding: utf-8 -*-
# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-collaboration specific hooks and operations"""

from cubicweb import Unauthorized
from cubicweb.server.hook import Hook, match_rtype
from cubicweb.predicates import adaptable


class GrantOwnersPermissions(Hook):
    """Grant read/write permissions to the owners of the collaborative
    entity."""
    __regid__ = 'collaboration.set_owners_permissions'
    __select__ = Hook.__select__ &  match_rtype('owned_by')
    events = ('after_add_relation',)

    def __call__(self):
        adapted = self._cw.entity_from_eid(self.eidfrom).cw_adapt_to('ICollaborationSecurity')
        if adapted:
            # Else the collaborative entity does not support security
            # relations.
            adapted.grant_permission(self.eidto, 'write')


class RevokeOwnersPermissions(Hook):
    """Revoke read and write permissions for users who do not own the collaborative
    entity anymore."""
    __regid__ = 'collaboration.revoke_owners_permissions'
    __select__ = Hook.__select__ &  match_rtype('owned_by')
    events = ('after_delete_relation',)

    def __call__(self):
        entity = self._cw.entity_from_eid(self.eidfrom)
        adapted = entity.cw_adapt_to('ICollaborationSecurity')
        if adapted:
            # Else the collaborative entity does not support security
            # relations.
            adapted.revoke_permission(self.eidto, 'write')
            c_adapted = entity.cw_adapt_to('ICollaborative')
            if self.eidto not in (user.eid for user in c_adapted.collaborators):
                adapted.revoke_permission(self.eidto, 'read')


class GrantReadCollaboratorsHook(Hook):
    __regid__ = 'grant-read-collaborators'
    __select__ = Hook.__select__ & match_rtype('collaborates_on')
    events = ('before_add_relation',)

    def __call__(self):
        adapted = self._cw.entity_from_eid(self.eidto).cw_adapt_to('ICollaborationSecurity')
        adapted.grant_permission(self.eidfrom, 'read')


class RevokeReadCollaboratorsHook(Hook):
    __regid__ = 'revoke-read-collaborators'
    __select__ = Hook.__select__ & match_rtype('collaborates_on')
    events = ('after_delete_relation',)

    def __call__(self):
        entity = self._cw.entity_from_eid(self.eidto)
        adapted = entity.cw_adapt_to('ICollaborationSecurity')
        if self.eidfrom not in (u.eid for u in entity.owned_by):
            adapted.revoke_permission(self.eidfrom, 'read')


